@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2006-2011 PLT Scheme Inc.
+  Copyright (c) 2006-2013 PLT Design Inc.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -70,6 +70,14 @@ typedef void *(*Get_Stack_Proc)();
 
 #define CACHE_STACK_MIN_TRIGGER 128
 
+/* Normally, caching of a native trace on the stack ensures that
+   `current-continuation-marks' is effectively constant-time. 
+   It's possible to have a deep stack section with no recognizable anchors
+   for caching, however, in which case `current-continuation-marks'
+   can become O(n); avoid that pathological case by limiting the
+   number of non-anchor frames in a row that we're willing to traverse. */
+#define UNKNOWN_FRAME_LIMIT 64
+
 #define USE_STACK_CHECK 0
 
 #if USE_STACK_CHECK
@@ -117,6 +125,34 @@ uintptr_t scheme_approx_sp()
   return p;
 }
 
+#if defined( _WIN64) && !defined(__MINGW32__)
+# define USE_WIN64_UNWIND
+# ifndef UNWIND_HISTORY_TABLE_SIZE
+extern PRUNTIME_FUNCTION WINAPI RtlLookupFunctionEntry(ULONG64, ULONG64*, void*);
+extern PVOID WINAPI RtlVirtualUnwind(DWORD, DWORD64, DWORD64, PRUNTIME_FUNCTION,
+				     PCONTEXT, PVOID, PDWORD64, PVOID);
+# endif
+#endif
+
+static void set_cache(void *p, Scheme_Object *last)
+{
+  int pos;
+
+  if (stack_cache_stack_pos >= (STACK_CACHE_SIZE - 1)) {
+    /* Make room on the stack */
+    void **z;
+    z = (void **)stack_cache_stack[stack_cache_stack_pos].stack_frame;
+    *z = stack_cache_stack[stack_cache_stack_pos].orig_return_address;
+    --stack_cache_stack_pos;
+  }
+
+  pos = (int)++stack_cache_stack_pos;
+  stack_cache_stack[pos].orig_return_address = ((void **)p)[RETURN_ADDRESS_OFFSET];
+  stack_cache_stack[pos].stack_frame = (void *)(((void **)p) + RETURN_ADDRESS_OFFSET);
+  stack_cache_stack[pos].cache = last;
+  ((void **)p)[RETURN_ADDRESS_OFFSET] = sjc.stack_cache_pop_code;
+}
+
 Scheme_Object *scheme_native_stack_trace(void)
 {
   void *p, *q;
@@ -134,6 +170,7 @@ Scheme_Object *scheme_native_stack_trace(void)
   int use_unw = 0;
   int shift_cache_to_next = 0;
   int added_list_elem;
+  int unsuccess = 0;
 
   if (!sjc.get_stack_pointer_code)
     return NULL;
@@ -180,7 +217,106 @@ Scheme_Object *scheme_native_stack_trace(void)
 #endif
   }
 
-  while (1) {
+#ifdef USE_WIN64_UNWIND
+  {
+    CONTEXT ctx;
+    PRUNTIME_FUNCTION rf;
+    ULONG64 base, ef;
+    void *data, *cache_sp = NULL;
+
+    RtlCaptureContext(&ctx);
+    while (unsuccess < UNKNOWN_FRAME_LIMIT) {
+      name = find_symbol((uintptr_t)ctx.Rip);
+      if (name) {
+	/* Unwind manually */
+	uintptr_t *fp = (uintptr_t *)ctx.Rbp;
+        if (!(STK_COMP((uintptr_t)fp, stack_end)
+              && STK_COMP(stack_start, (uintptr_t)fp))) {
+          /* out-of-range frame pointer; give up */
+          break;
+        } else if (SCHEME_FALSEP(name) || SCHEME_VOIDP(name)) {
+	  /* "quick" call convention */
+	  if (SCHEME_VOIDP(name)) {
+	    /* JIT_LOCAL2 has the next return address */
+	    ctx.Rip = fp[JIT_LOCAL2 >> JIT_LOG_WORD_SIZE];
+	  } else {
+	    /* Push after local stack of return-address proc
+	       may have the next return address */
+	    ctx.Rip = fp[-(3 + LOCAL_FRAME_SIZE + 1)];
+	  }
+	  name = find_symbol((uintptr_t)ctx.Rip);
+	} else {
+	  /* normal JIT function convention */
+	}
+
+	cache_sp = (void *)fp;
+
+	if (SCHEME_EOFP(name)) {
+	  /* JIT_LOCAL2 has the name to use */
+	  name = *(Scheme_Object **)fp[JIT_LOCAL2 >> JIT_LOG_WORD_SIZE];
+	}
+
+	ctx.Rsp = ctx.Rbp + (2 * sizeof(void*));
+# ifdef NEED_LOCAL4
+	ctx.R14 = fp[-JIT_LOCAL4_OFFSET];
+# endif
+	ctx.Rbp = fp[0];
+	ctx.Rbx = fp[-1];
+	ctx.Rsi = fp[-2];
+	ctx.Rdi = fp[-3];
+	ctx.Rip = fp[1];
+
+	if (SCHEME_NULLP(name))
+	  name = NULL;
+      } else {
+	unsuccess++;
+	rf = RtlLookupFunctionEntry(ctx.Rip, &base, NULL);
+	if (rf) {
+	  RtlVirtualUnwind(0x0, base, ctx.Rip,
+			   rf, &ctx, &data, &ef, NULL);
+	} else {
+	  break;
+	}
+      }
+
+      if (name) {
+	name = scheme_make_pair(name, scheme_null);
+	if (last)
+	  SCHEME_CDR(last) = name;
+	else
+	  first = name;
+	last = name;
+      }
+
+      if (cache_sp) {
+	if (STK_COMP((uintptr_t)halfway, (uintptr_t)cache_sp)) {
+	  set_cache(cache_sp, last);
+	  halfway = stack_end;
+	  unsuccess = -100000; /* if we got halfway, no need to bail out later */
+	}
+	cache_sp = NULL;
+      }
+
+      if (!(STK_COMP((uintptr_t)ctx.Rsp, stack_end)
+	    && STK_COMP(stack_start, (uintptr_t)ctx.Rsp))) {
+	/* out of stack range */
+	break;
+      }
+    }
+
+    if (last)
+      SCHEME_CDR(last) = tail;
+    else
+      first = tail;
+    
+    if (SCHEME_NULLP(first))
+      return NULL;
+    else
+      return first;
+  }
+#endif
+
+  while (unsuccess < UNKNOWN_FRAME_LIMIT) {
 #ifdef MZ_USE_DWARF_LIBUNWIND
     if (use_unw) {
       q = (void *)unw_get_ip(&c);
@@ -270,9 +406,14 @@ Scheme_Object *scheme_native_stack_trace(void)
         shift_cache_to_next = 0;
       }
       added_list_elem = 1;
-    } else 
+    } else
       added_list_elem = 0;
 
+    if (!name)
+      unsuccess++;
+    else
+      unsuccess = 0;
+
     /* Cache the result halfway up the stack, if possible. Only cache
        on frames where the previous frame had a return address with a
        name, because an arbitrary frame's return address on the stack
@@ -281,25 +422,11 @@ Scheme_Object *scheme_native_stack_trace(void)
        it will use the return address from the stack. */
     if (STK_COMP((uintptr_t)halfway, (uintptr_t)p)
 	&& prev_had_name) {
-      int pos;
-
-      if (stack_cache_stack_pos >= (STACK_CACHE_SIZE - 1)) {
-	/* Make room on the stack */
-	void **z;
-	z = (void **)stack_cache_stack[stack_cache_stack_pos].stack_frame;
-	*z = stack_cache_stack[stack_cache_stack_pos].orig_return_address;
-	--stack_cache_stack_pos;
-      }
-
-      pos = ++stack_cache_stack_pos;
-      stack_cache_stack[pos].orig_return_address = ((void **)p)[RETURN_ADDRESS_OFFSET];
-      stack_cache_stack[pos].stack_frame = (void *)(((void **)p) + RETURN_ADDRESS_OFFSET);
-      stack_cache_stack[pos].cache = last;
-      ((void **)p)[RETURN_ADDRESS_OFFSET] = sjc.stack_cache_pop_code;
+      set_cache(p, last);
       if (!added_list_elem)
-        shift_cache_to_next = 1;
-
+	shift_cache_to_next = 1;
       halfway = stack_end;
+      unsuccess = -100000; /* if we got halfway, no need to bail out later */
     }
 
     prev_had_name = !!name;
